# include <stdlib.h>
# include <assert.h>
# include "gcc-externs.h"
# include "sdg.h"
# include "bb.h"
# include "regset.ic"

/* Size of the set(in # of DEFN_SET_ELT_TYPE elts) to represent all defns. */
int defnSetSize;

	/* For the 'bb' being processed, keeps track of the ids of defns
	 * within the basic block */
int defnId;

CFG cfg; /* The control flow graph */

extern int 		   numDefns;	/* Number of definitions in this function */
extern BasicBlock *currBB;		/* Current basic block being processed */

void
AddCfgEdge(int pred, int succ)
{
/* Add 'succ' as a successor of 'pred'
 * Add 'pred' as a predecessor of 'succ' */

	BasicBlock *predBB = GetBasicBlock(pred);
	BasicBlock *succBB = GetBasicBlock(succ);

	Edge *edge = (Edge *)malloc(sizeof(Edge));
	edge->isBackEdge = FALSE;
	edge->isCycleEdge = FALSE;
	edge->src = predBB;
	edge->dst = succBB;

	Insert(&predBB->succs, edge);
	Insert(&succBB->preds, edge);
}

void
SetControlFlowEdges(BasicBlock *bb)
{
/*
 * NOTE : If this code can be migrated to '../gcc-2.7.1/flow.c' then
 * some of the information like 'uid_block_number' can be thrown away
 * in '../gcc-2.7.1/flow.c'.  Also, the information about (ii) below
 * is available there.  Can do this later ..
 * Also, some of this code like 'Get first non-trivial instr' can be
 * removed and functions from 'rtl.c' can be used instead which provide
 * the same functionality.
 */

/*
 *   (i) If 'bb' has a label, then possibly, other bbs can jump to this one
 *        So, firstly find out if 'bb' has a label and if so, identify bbs
 *        which jump to 'bb' and add cfg edges
 *  (ii) Secondly, if 'bb' has no jump instruction, then 'bb' falls thru
 *        to the succeeding basic block.  Add an edge in that case
 * (iii) Thirdly, if 'bb' has a jump instruction, and it turns out to be
 *        a conditional branch, even then 'bb' can fall thru to the next
 *        basic block.
 */

#define BLOCK_NUM(INSN)  uid_block_number[INSN_UID (INSN)]

	int         bbId = bb->id;
	rtx         head = basic_block_head[bbId];
	rtx			end = basic_block_end[bbId];
	rtx         insn, tmp, label;
	rtx         jump;
	RTX_CODE    code;

	label = NULL;

		/* Get the first non-trivial instruction */
	tmp = head;
	do {
		insn = tmp;
		tmp = NEXT_INSN(insn);
		code = GET_CODE(insn);
			/* Since any bb has only 1 label, this is fine */
		if (code == CODE_LABEL)
			label = insn;
		if (code == INSN || code == CALL_INSN || code == JUMP_INSN)
			break;
	} while (insn != end);

		/* Got this for-loop from 'gcc'(flow.c) */
	if (label != NULL)
		for (jump = LABEL_REFS(head); jump != head; jump = LABEL_NEXTREF(jump))
			AddCfgEdge(BLOCK_NUM(CONTAINING_INSN(jump)), bbId);

		/* Get the last non-trivial instruction */
	tmp = end;
	do {
		insn = tmp;
		tmp = PREV_INSN(insn);
		code = GET_CODE(insn);
		if (code == INSN || code == CALL_INSN || code == JUMP_INSN)
			break;
	} while (insn != head);

	if ((code != JUMP_INSN) || (condjump_p(insn) && !simplejump_p(insn))) {
		if (bbId < n_basic_blocks-1)
			AddCfgEdge(bbId, bbId+1);
	}
}

void
InitBasicBlock(int bbId)
{
/* Initialise the basicblock 'bb' */
	BasicBlock *bb = GetBasicBlock(bbId);
	bb->id = bbId;
	bb->prob = -1.0;
	Init(bb->succs);
	Init(bb->preds);
	Init(bb->activeRegs);
	bb->startDefnId = 0;
	bb->startNode = NULL;
	bb->innerLoop = NULL;
	bb->dummyExit = NULL;
	bb->loopDepth = 0;
	bb->reachable = 0;
}

void
AllocMemory(BasicBlock *bb)
{
/* Allocate memory for the pregs defined/live in this basic block
 * Needed for computing reaching defns. */

	int      bbId = bb->id;
	DefnSet  mem;
	Preg    *pregs;
	int      i, count;
	int      numActiveRegs = 0;

# if 0
	fprintf(stderr, "Active regs for basicBlock %d : ", bbId);
# endif

		/* Count the # of active pseudo regs in the current bb */
	for (i = 0; i < max_regno; i++) {
				/* Dont bother to track stack_pointer/frame_pointer */
		if (i == FRAME_POINTER_REGNUM || i == STACK_POINTER_REGNUM)	
			continue;

		if (   BitIsSet(basic_block_live_at_start[bbId], i)
			|| BitIsSet(regsDefinedInBB[bbId], i))
		{
# if 0
			fprintf(stderr, "%d ", i);
# endif
			numActiveRegs++;
		}
	}

# if 0
	fputc('\n', stderr,);
# endif

		/* For each active reg., 3 defn. sets are required
		 * (i) defsIn (ii) defsOut (iii) tmp.  The size of
		 * this defn. set is decided by 'numDefns'
		 */
	pregs = (Preg *)calloc(numActiveRegs, sizeof(Preg));
	mem = (DefnSet)malloc(numActiveRegs*3*defnSetSize*BYTES_PER_DEFN_SET_ELT);
	bzero((unsigned char *)mem, numActiveRegs*3*defnSetSize*BYTES_PER_DEFN_SET_ELT);
	count = 0;
	for (i = 0; i < max_regno; i++) {
				/* Dont bother to track stack_pointer/frame_pointer */
		if (i == FRAME_POINTER_REGNUM || i == STACK_POINTER_REGNUM)
			continue;

		if (   BitIsSet(basic_block_live_at_start[bbId], i)
			|| BitIsSet(regsDefinedInBB[bbId], i))
		{
			pregs[count].regNo = i;
			pregs[count].defsIn = mem;
			mem += defnSetSize;
			pregs[count].defsOut = mem;
			mem += defnSetSize;
			pregs[count].tmp = mem;
			mem += defnSetSize;

				/* Append this preg to the list of active pregs */
			Append(&bb->activeRegs, &pregs[count]);
			count++;
		}
	}
}

Preg *
GetPreg(BasicBlock *bb, int regNo)
{
	Iterator  iter;
	Preg     *reg;

	reg = (Preg *)InitIter(iter, bb->activeRegs);
	while (reg != NULL) {
		if (reg->regNo == regNo)
			return reg;

		reg = (Preg *)GetNext(iter);
	}

	return NULL;
}

void
Defn_Union(DefnSet set1, DefnSet set2)
{
	int i;

	for (i = 0; i < defnSetSize; i++)
		set1[i] |= set2[i];
}

void
ComputeDefsIn(BasicBlock *bb)
{
/* For each of the active registers, compute 'defsIn' 
 * as the union of the 'defsOut' of bb's predecessors */

	Iterator    iter1, iter2;
	Edge       *edge;
	BasicBlock *pred;
	Preg       *reg, *tmpReg;

	reg = (Preg *)InitIter(iter1, bb->activeRegs);
	while (reg != NULL) {

/* No need to clear the 'defsIn' set because if in an earlier pass,
 * some defn. reached the beginning of 'bb', it will still reach 
 * the beginning of 'bb' in this pass */

				/* Go thru' the list of bb's predecessors and
				 * compute the union of 'defsOut' of these bbs */
		for (edge = (Edge *)InitIter(iter2, bb->preds);
		     edge != NULL;
			 edge = (Edge *)GetNext(iter2))
		{
				/* It may happen that 'reg' is not active
				 * in 'pred' in which case 'tmpReg' will be NULL */
			pred = edge->src;
			tmpReg = GetPreg(pred, reg->regNo);
			if (tmpReg != NULL)
				Defn_Union(reg->defsIn, tmpReg->defsOut);
		}

		reg = (Preg *)GetNext(iter1);
	}
}

int
CopyIfNotSame(DefnSet oldSet, DefnSet newSet)
{
/* Check if 'oldSet' and 'newSet' are the same.
 * If they are, then return 0, 
 * else copy newSet to oldSet and return 1
 */

	int i;

	for (i = 0; i < defnSetSize; i++)
		if (oldSet[i] != newSet[i]) {
			bcopy((unsigned char *)&newSet[i], (unsigned char *)&oldSet[i], 
					(defnSetSize - i)*BYTES_PER_DEFN_SET_ELT);
			return 1;
		}

	return 0;
}

void
UpdateReachingDefnInCurrBB(int regNum)
{
/* 'regNum' has been written into.
 * Update 'tmp' for this preg in the current basic block
 * to reflect this new defn.  */

	if (regNum != STACK_POINTER_REGNUM && regNum != FRAME_POINTER_REGNUM) {
		Preg *reg = GetPreg(currBB, regNum);
		Defn_ClearSet(reg->tmp);
		Defn_SetBit(reg->tmp, defnId);
		defnId++;
	}
}

void
Handle_PARALLEL_rtx(rtx insn)
{
/* PARALLEL holds a vector of expressions each of which is a side-effect expression.  
 * The side-effects do not take place until the 'RHS' of all exprs. in the
 * vector are completed.  But, we can update reaching defn info immediately
 * since we are only using this info. to compute 'DefsOut' */

	int  i;
	rtx  expr, dest;
	extern void HandleInstr(rtx);

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		expr = XVECEXP(insn, 0, i);
		if (GET_CODE(expr) == SET) {
			dest = SET_DEST(expr);
			switch (GET_CODE(dest)) {
				case REG :
					UpdateReachingDefnInCurrBB(REGNO(dest));
					break;

				case SUBREG :
					UpdateReachingDefnInCurrBB(REGNO(SUBREG_REG(dest)));
					break;

				case STRICT_LOW_PART :
					UpdateReachingDefnInCurrBB(REGNO(SUBREG_REG(XEXP(dest, 0))));
					break;
				
				default :
					break;
			}
		}
	}
}

void
Handle_SEQUENCE_rtx(rtx insn)
{
/* SEQUENCE holds a vector of instructions */

	int 	 i;
	rtx      instr;
	RTX_CODE code;
	extern void HandleInstr(rtx);

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		instr = XVECEXP(insn, 0, i);
		code = GET_CODE(instr);
		if (code == INSN || code == CALL_INSN)
			HandleInstr(PATTERN(instr));
	}
}

void
Handle_UNSPEC_rtx(rtx insn)
{
/* UNSPEC/UNSPEC_VOLATILE have operands which're used in some
 * m/c specific operation(s).  These are considered to define/use
 * all their operands */

	int   i;
	rtx   elt;

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		elt = XVECEXP(insn, 0, i);
		if (REG_P(elt))
			UpdateReachingDefnInCurrBB(REGNO(elt));
	}
}

void
HandleInstr(rtx insn)
{
	rtx    dest;

	switch (GET_CODE(insn)) {
	case RETURN :
	case USE :
	case ADDR_VEC :
	case ADDR_DIFF_VEC :
	case CALL :
			/* In all these cases, there is no new defn. Continue ... */
		break;

	case TRAP_IF :
		break;

	case ASM_INPUT :
		fprintf(stderr, "Inline ASM not yet handled...\n");
		abort();

	case PARALLEL :
		Handle_PARALLEL_rtx(insn);
		break;

	case SEQUENCE :
		Handle_SEQUENCE_rtx(insn);
		break;

	case CLOBBER :
		dest = XEXP(insn, 0);
		if (REG_P(dest))
			UpdateReachingDefnInCurrBB(REGNO(dest));
		break;

	case SET :
		dest = SET_DEST(insn);
		switch (GET_CODE(dest)) {
			case REG :
				UpdateReachingDefnInCurrBB(REGNO(dest));
				break;

			case SUBREG :
				UpdateReachingDefnInCurrBB(REGNO(SUBREG_REG(dest)));
				break;

			case STRICT_LOW_PART :
				UpdateReachingDefnInCurrBB(REGNO(SUBREG_REG(XEXP(dest, 0))));
				break;
			
			default :
				break;
		}
		break;

	case UNSPEC :
	case UNSPEC_VOLATILE :
			/* These are considered to set/use all its operands */
		Handle_UNSPEC_rtx(insn);
		break;

	default :
		fprintf(stderr, "Unhandled case in \'HandleInstr\'\n");
		abort();
	}
}

int
ComputeDefsOut(BasicBlock *bb)
{
/* Compute 'defsOut' for each of the active registers
 * Compute this in 'tmp' so that at the end, we can compare
 * the two to see if 'defsOut' has changed.  If so, we need
 * return 1 indicating that another pass is needed */

	int 	  changed = 0;
	Preg     *reg;
	Iterator  iter;
	rtx       first = basic_block_head[bb->id];
	rtx       last = basic_block_end[bb->id];
	rtx 	  insn, next;
	RTX_CODE  code;

		/* Set 'currBB' */
	currBB = bb;

		/* Needed to keep track of the definitions within this bb */
	defnId = bb->startDefnId;

		/* Initially, copy 'defsIn' into 'tmp' */
	reg = (Preg *)InitIter(iter, bb->activeRegs);
	while (reg != NULL) {
		Defn_CopySet(reg->defsIn, reg->tmp);
		reg = (Preg *)GetNext(iter);
	}

	for (insn = first; ; insn = next) {
		next = NEXT_INSN(insn);
		code = GET_CODE(insn);

			/* Notes, barriers and labels are useless... */
			/* So are jumps.  They cant result in any new defns. */
			/* For INSN or CALL_INSN, identify possible defns. */
		if (code == INSN || code == CALL_INSN)
			HandleInstr(PATTERN(insn));

		if (insn == last)
			break;
	}

		/* Now, compare 'tmp' and 'defsOut' and check if they
		 * are different.  Update 'defsOut' in those cases */
	reg = (Preg *)InitIter(iter, bb->activeRegs);
	while (reg != NULL) {
		if (CopyIfNotSame(reg->defsOut, reg->tmp))
			changed = 1;
		reg = (Preg *)GetNext(iter);
	}

	return changed;
}

# ifndef PLAIN_SLICING
unsigned long *profile_BBcount;

void
ReadInBBexecutionCounts()
{
	int   i;
	FILE *fp;
	char  buffer1[100], buffer2[100];
	int   numBlocks, dummy;
	char *profileFileName = "bb.out";

	profile_BBcount = (unsigned long *)malloc(sizeof(unsigned long) * n_basic_blocks);
	for (i = 0; i < n_basic_blocks; i++)
		profile_BBcount[i] = 0;

	fp = fopen(profileFileName, "r");
	if (fp == NULL) {
		fprintf(stderr, "Cannot open profile file \'%s\'\n", profileFileName);
		return;
	}

	while (1) {
		if (fscanf(fp, "%s %s", buffer1, buffer2) != 2)
			break;

		if (   (strcmp(buffer1, dump_base_name) == 0)
			&& (strcmp(buffer2, currentFnName) == 0))
		{
			fscanf(fp, "%d", &numBlocks);
			if (numBlocks != n_basic_blocks) {
				fprintf(stderr, "ERROR :\n");
				fprintf(stderr, "CURRENT numBBs = %d\n", n_basic_blocks);
				fprintf(stderr, "PROFILE numBBs = %d\n", numBlocks);
			}

			for (i = 0; i < numBlocks; i++)
				fscanf(fp, "%uld", &profile_BBcount[i]);

			break;
		}
		else {
			fscanf(fp, "%d", &numBlocks);
			for (i = 0; i < numBlocks; i++)
				fscanf(fp, "%d", &dummy);
		}

		if (feof(fp))
			break;
	}

	fclose(fp);
}
# endif

void
AddSuccessorsToWorkList(BasicBlock *bb, List *workList, BBset set)
{
	Iterator    iter;
	BasicBlock *succ;
	Edge       *edge;

	for (edge = (Edge *)InitIter(iter, bb->succs);
		 edge != NULL;
		 edge = (Edge *)GetNext(iter))
	{
		succ = edge->dst;
				/* Avoid duplicates in the work list */
		if (!BB_BitIsSet(set, succ->id)) {
			Append(workList, succ);
			BB_SetBit(set, succ->id);
		}
	}
}

void
FreeCfg()
{
	int 		i;
	BasicBlock *bb, *dummyExit;
	Preg       *reg;

	for (i = 0; i < n_basic_blocks; i++) {
		bb = GetBasicBlock(i);
		FreeList(&bb->preds);
		FreeList(&bb->succs);
		    /*
			 * The pregs are allocated in a single chunk starting at
			 * the first preg. All mem. for defsIn, defsOut is also 
			 * allocated as a single chunk
			 */
		if (!IsEmpty(bb->activeRegs)) {
			reg = GetHead(bb->activeRegs);
			free(reg->defsIn);
			free(reg);
			FreeList(&bb->activeRegs);
		}

		dummyExit = bb->dummyExit;
		while (dummyExit != NULL) {
			bb = dummyExit->dummyExit;
			free(dummyExit);
			dummyExit = bb;
		}
	}

	free(cfg);
}

void
PrintCfgEdges(BasicBlock *bb)
{
	Iterator    iter;
	BasicBlock *tmp;
	Edge       *edge;
	extern FILE *sdg_dump_file;

	fprintf(sdg_dump_file, "######### Basic block %d #########\n", bb->id);
	fprintf(sdg_dump_file, "Prob       : %1.3f\n", bb->prob);
	fprintf(sdg_dump_file, "Loop depth : %d\n", bb->loopDepth);
	if (bb->loopDepth == 0)
		fprintf(sdg_dump_file, "Loop hdr id: -- NOT IN ANY LOOP --\n");
	else
		fprintf(sdg_dump_file, "Loop hdr id: %d\n", bb->innerLoop->hdr->id);
	fprintf(sdg_dump_file, "First insn : %d\nLast insn  : %d\n",
			INSN_UID(basic_block_head[bb->id]),
			INSN_UID(basic_block_end[bb->id]));
	fprintf(sdg_dump_file, "Preds      : ");
	edge = (Edge *)InitIter(iter, bb->preds); 
	while (edge != NULL) {
		fprintf(sdg_dump_file, "%d ", edge->src->id);
		edge = (Edge *)GetNext(iter);
	}
	fprintf(sdg_dump_file, "\nSuccs      : ");
	edge = (Edge *)InitIter(iter, bb->succs); 
	while (edge != NULL) {
		fprintf(sdg_dump_file, "%d ", edge->dst->id);
		edge = (Edge *)GetNext(iter);
	}
	fprintf(sdg_dump_file, "\n");
}

# if 0

void
PrintDefsIn(BasicBlock *bb)
{
	Iterator  iter;
	Preg     *reg;
	int       i;

	fprintf(stderr, "##### BasicBlock %d : #####\n", bb->id);
	reg = (Preg *)InitIter(iter, bb->activeRegs);
	while (reg != NULL) {
		fprintf(stderr, "*** Reg %d : ***\n", reg->regNo);
		for (i = 0; i < numDefns; i++)
			if (Defn_BitIsSet(reg->defsIn, i)) 
				fprintf(stderr, "%d ", i);
		fputc('\n', stderr);

		reg = (Preg *)GetNext(iter);
	}
}

# endif
