# ifndef __BASIC_BLOCK__
# define __BASIC_BLOCK__

# include "types.h"

# include "bb.ic"

		/* Size of the defn set */
extern int defnSetSize;

		/* Id. of the latest defn. encountered */
extern int defnId;

typedef struct {
	int   	 regNo;		/* Pseudo Reg # */
	DefnSet  defsIn;	/* For a bb, the defns. of this preg
						 * which reach the start of the bb */
	DefnSet  defsOut;	/* For a bb, the defns. of this preg which
						 * reach the end of the bb. */
	DefnSet  tmp;		/* A temp reaching defn. set needed
						 * while processing a bb */
} Preg;

struct BasicBlock {
	int      id;			/* The id of this basic block */
	short    reachable : 1; /* Can this bb be reached from entry? */
	float    prob;			/* Probability of this basic block executing */
	float    succProb;    	/* Probability of each succ. being visited
							 * For the present, this is 'prob/numSuccs' */
	short    loopDepth;		/* Loop nesting level.  0 => Top level */
	Loop    *innerLoop;		/* Pointer to the innermost loop that contains
							 * this basic block */
	BasicBlock *dummyExit;  /* If this bb takes us out of a loop, then in
							 * certain cases, a dummy exit would be necessary */
	SdgNode *startNode;		/* The first node of the sdg corr. to the
							 * first instr. of this basic block */
	int      startDefnId; 	/* Id of the first definition in this bb */
	List     preds;			/* The list of predecessor bbs */
	List     succs;			/* The list of successor bbs */
	List     activeRegs;	/* All regs. which are active(defined/live) 
							 * in this basic block */
};

typedef struct {
	BasicBlock *src;		/* Edge source */
	BasicBlock *dst;		/* Edge destination */
	short       isBackEdge; /* Is this a back edge? */
	short       isCycleEdge;/* Is this edge part of a non-backedge cycle? */
} Edge;

	/* Assumes a single-entry multiple-exit loop */
struct Loop {
	Loop       *enclosingLoop;  /* If this loop is nested inside another loop,
								 * the innermost such loop */
	List        body;			/* Loop body - list of basic blocks */
	BasicBlock *hdr;			/* Loop header */
	BBset       backEdges;		/* Mask of backedges of this loop */
	BBset       visitMask;		/* Mask of backedges that've been visited */
	float       backEdgeProb;   /* Sum of probabilities of traversing 
							     * the back edges */
};

			/* The control flow graph */
extern CFG         cfg;
extern BasicBlock *currBB;
extern List        loopList;

extern void  InitBasicBlock(int bbId);
extern void  AllocMemory(BasicBlock *bb);
extern void  ComputeBBexecutionProbabilities();
extern void SetControlFlowEdges(BasicBlock *bb);
extern Preg *GetPreg(BasicBlock *bb, int regNo);
extern void  ComputeDefsIn(BasicBlock *bb);
extern int   ComputeDefsOut(BasicBlock *bb);
extern void  AddSuccessorsToWorkList(BasicBlock *bb, List *list, BBset);
extern void  IdentifyDependenciesInBB(BasicBlock *bb);
extern void  FreeCfg();
extern void  UpdateReachingDefnInCurrBB(int regNum);
# if 1
extern void PrintDefsIn(BasicBlock *);
extern void PrintCfgEdges(BasicBlock *);
# endif

# endif __BASIC_BLOCK__
