# include <assert.h>
# include <stdio.h>
# include "gcc-externs.h"

static int replaceSrc;
static int replaceDst;

void
ReplaceReg(rtx expr, rtx *exprPtr, rtx oldReg, rtx newReg)
{
	void ReplaceRegInExpr(rtx expr, rtx oldReg, rtx newReg);
	
	if (GET_CODE(expr) != REG)
		ReplaceRegInExpr(expr, oldReg, newReg);
	else if (expr == oldReg)
		*exprPtr = newReg;
}

void 
ReplaceRegInExpr(rtx expr, rtx oldReg, rtx newReg)
{
	RTX_CODE  code;
	int       regNum;
	rtx       op1, op2;
	
	switch (code = GET_CODE(expr)) {
			/* The following cases are constant expressions
			 * which do not use any pseudo regs. */
			/* CHECK : Verify that the above is true for 'CONST_DOUBLE' */
	case CONST_INT :
	case CONST_STRING :
	case SYMBOL_REF :
	case LABEL_REF :
	case CONST :
	case HIGH :
	case CONST_DOUBLE :
		break;

			/* For these operand types, there is no 
			 * use of a pseudo reg => no dependency edge added. */
	case CC0 :
	case PC :
		break;

			/* Nothing to do.  No pseudo regs used */
	case SCRATCH :
		break;

			/* Call.  Nothing to do.  No pseudo regs. used */
	case CALL :
		break;

			/* Cond jump. */
	case IF_THEN_ELSE :
		ReplaceReg(XEXP(expr, 0), &XEXP(expr, 0), oldReg, newReg); /* cond */
		ReplaceReg(XEXP(expr, 1), &XEXP(expr, 1), oldReg, newReg); /* then */
		ReplaceReg(XEXP(expr, 2), &XEXP(expr, 2), oldReg, newReg); /* else */
		break;

	case COND :
		fprintf(stderr, 
				"COND is not supposed to appear as an instr. pattern\n");
		abort();
		break;

	case REG :
			/* Cannot come here .. */
		assert(0);
		break;

	case SUBREG :
		ReplaceReg(SUBREG_REG(expr), &SUBREG_REG(expr), oldReg, newReg);
		break;

	case STRICT_LOW_PART :
		ReplaceReg(SUBREG_REG(XEXP(expr, 0)), &SUBREG_REG(XEXP(expr, 0)), oldReg, newReg);
		break;

				/* Binary arithmetic operations */
	case PLUS :
	case LO_SUM :
	case MINUS :
	case COMPARE :
	case MULT :
	case DIV : 
	case UDIV :
	case MOD : 
	case UMOD :
	case SMIN :
	case SMAX :
	case UMIN :
	case UMAX :
				/* Binary logical and bit operations */
	case AND :
	case IOR :
	case XOR :
	case ASHIFT :
	case LSHIFTRT :
	case ASHIFTRT :
	case ROTATE :
	case ROTATERT :
				/* Binary comparison operations */
	case EQ :
	case NE :
	case GT :
	case GTU :
	case LT :
	case LTU :
	case GE :
	case GEU :
	case LE :
	case LEU :
		ReplaceReg(XEXP(expr, 0), &XEXP(expr, 0), oldReg, newReg);
		ReplaceReg(XEXP(expr, 1), &XEXP(expr, 1), oldReg, newReg);
		break;

				/* Unary arithmetic operations */
	case NOT :
	case NEG :
	case ABS :
	case SQRT :
	case FFS :
				/* Bit-field operations  */
	case SIGN_EXTRACT :
	case ZERO_EXTRACT :
				/* Conversions */
	case FIX :
	case UNSIGNED_FIX :
	case FLOAT_EXTEND :
	case FLOAT_TRUNCATE :
	case SIGN_EXTEND :
	case ZERO_EXTEND :
	case FLOAT :
	case UNSIGNED_FLOAT :
	case TRUNCATE :
				/* Memory address */
	case MEM :
		ReplaceReg(XEXP(expr, 0), &XEXP(expr, 0), oldReg, newReg); 
		break;

			/* Unhandled codes */
	default :
		fprintf(stderr, "Unhandled case in \'ReplaceRegInExpr\'\n");
		abort();
	}
}

void
ReplaceRegIn_SET_rtx(rtx insn, rtx oldReg, rtx newReg)
{
	if (replaceSrc)
		ReplaceReg(SET_SRC(insn), &SET_SRC(insn), oldReg, newReg);
	if (replaceDst)
		ReplaceReg(SET_DEST(insn), &SET_DEST(insn), oldReg, newReg);
}

void
ReplaceRegIn_SEQUENCE_rtx(rtx insn, rtx oldReg, rtx newReg)
{
/* SEQUENCE holds a vector of instructions */

	int 	 i;
	rtx      instr;
	RTX_CODE code;
	extern void ReplaceRegInInstr(rtx, rtx oldReg, rtx newReg);

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		instr = XVECEXP(insn, 0, i);
		code = GET_CODE(instr);
		if (code == INSN || code == CALL_INSN)
			ReplaceRegInInstr(instr, oldReg, newReg);
	}
}

void
ReplaceRegIn_PARALLEL_rtx(rtx insn, rtx oldReg, rtx newReg)
{
/* PARALLEL holds a vector of expressions each
 * of which is a side-effect expression */

	int   i;
	rtx   elt;

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		elt = XVECEXP(insn, 0, i);
		if (GET_CODE(elt) == SET)
			ReplaceRegIn_SET_rtx(elt, oldReg, newReg);
	}
}

static void
Replace(rtx insn, rtx oldReg, rtx newReg)
{
/* Change the mode of all registers in this instruction
 * to 'SF' mode */

	rtx   body, dest;

	switch (GET_CODE(insn)) {
	case NOTE :
	case BARRIER :
	case CODE_LABEL :
		break;
		
	case INSN :
	case JUMP_INSN :
	case CALL_INSN :
		Replace(PATTERN(insn), oldReg, newReg);
		break;

	case ADDR_VEC :
	case ADDR_DIFF_VEC :
	case RETURN :
	case CALL :
	case USE :
	case CLOBBER :
	case TRAP_IF : /* Not yet handled */
	case ASM_INPUT :
	case UNSPEC :
	case UNSPEC_VOLATILE :
		break;

	case PARALLEL :
		ReplaceRegIn_PARALLEL_rtx(insn, oldReg, newReg);
		break;

	case SEQUENCE :
		ReplaceRegIn_SEQUENCE_rtx(insn, oldReg, newReg);
		break;

	case SET :
		ReplaceRegIn_SET_rtx(insn, oldReg, newReg);
		break;

	default :
		fprintf(stderr, "Unhandled case in \'Replace\'\n");
		abort();
	}
}

void
ReplaceDestRegInInstr(rtx insn, rtx oldReg, rtx newReg)
{
	replaceDst = 1;
	replaceSrc = 0;
	Replace(insn, oldReg, newReg);
}

void
ReplaceSourceRegInInstr(rtx insn, rtx oldReg, rtx newReg)
{
	replaceDst = 0;
	replaceSrc = 1;
	Replace(insn, oldReg, newReg);
}

void
ReplaceRegInInstr(rtx insn, rtx oldReg, rtx newReg)
{
	replaceDst = 1;
	replaceSrc = 1;
	Replace(insn, oldReg, newReg);
}
